/**
 * Copyright (c) 2014-2018 Enzien Audio Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _HEAVY_CONTROL_UNOP_H_
#define _HEAVY_CONTROL_UNOP_H_

#include "HvHeavyInternal.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef enum UnopType {
  HV_UNOP_ASIN,
  HV_UNOP_ASINH,
  HV_UNOP_ACOS,
  HV_UNOP_ACOSH,
  HV_UNOP_ATAN,
  HV_UNOP_ATANH,
  HV_UNOP_SIN,
  HV_UNOP_SINH,
  HV_UNOP_COS,
  HV_UNOP_COSH,
  HV_UNOP_TAN,
  HV_UNOP_TANH,
  HV_UNOP_EXP,
  HV_UNOP_ABS,
  HV_UNOP_SQRT,
  HV_UNOP_LOG,
  HV_UNOP_LOG2,
  HV_UNOP_LOG10,
  HV_UNOP_CEIL,
  HV_UNOP_FLOOR,
  HV_UNOP_ROUND
} UnopType;

void cUnop_onMessage(HeavyContextInterface *_c, UnopType op, const struct HvMessage *m,
    void (*sendMessage)(HeavyContextInterface *, int, const struct HvMessage *));

#ifdef __cplusplus
} // extern "C"
#endif

#endif // _HEAVY_CONTROL_UNOP_H_
