/**
 * Copyright (c) 2014-2018 Enzien Audio Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "HvControlCast.h"

void cCast_onMessage(HeavyContextInterface *_c, CastType castType, int letIn, const HvMessage *m,
    void (*sendMessage)(HeavyContextInterface *, int, const HvMessage *)) {
  switch (castType) {
    case HV_CAST_BANG: {
      HvMessage *n = HV_MESSAGE_ON_STACK(1);
      msg_initWithBang(n, msg_getTimestamp(m));
      sendMessage(_c, 0, n);
      break;
    }
    case HV_CAST_FLOAT: {
      if (msg_isFloat(m, 0)) {
        HvMessage *n = HV_MESSAGE_ON_STACK(1);
        msg_initWithFloat(n, msg_getTimestamp(m), msg_getFloat(m, 0));
        sendMessage(_c, 0, n);
      }
      break;
    }
    case HV_CAST_SYMBOL: {
      switch (msg_getType(m, 0)) {
        case HV_MSG_BANG: {
          HvMessage *n = HV_MESSAGE_ON_STACK(1);
          msg_initWithSymbol(n, msg_getTimestamp(m), "bang");
          sendMessage(_c, 0, n);
          break;
        }
        case HV_MSG_FLOAT: {
          HvMessage *n = HV_MESSAGE_ON_STACK(1);
          msg_initWithSymbol(n, msg_getTimestamp(m), "float");
          sendMessage(_c, 0, n);
          break;
        }
        case HV_MSG_SYMBOL: {
          sendMessage(_c, 0, m);
          break;
        }
        default: return;
      }
      break;
    }
    default: return;
  }
}
