{{copyright}}

#ifndef _HV_{{name|upper}}_WWISE_SOURCE_PLUGIN_FACTORY_H_
#define _HV_{{name|upper}}_WWISE_SOURCE_PLUGIN_FACTORY_H_

#ifndef AK_IOS
#include <AK/SoundEngine/Common/IAkPlugin.h>
#endif

AK_STATIC_LINK_PLUGIN(Hv_{{name}}_Wwise{{plugin_type}}Plugin);

#endif // _HV_{{name|upper}}_WWISE_SOURCE_PLUGIN_FACTORY_H_
{# force new line #}
